@@ -17,9 +17,11 @@
 
 from __future__ import division, print_function, absolute_import
 
+import sys
 import enum
 import math
 import random
+import string
 import datetime as dt
 import operator
 from uuid import UUID
@@ -27,22 +29,27 @@ from decimal import Context, Decimal, localcontext
 from inspect import isclass, isfunction
 from fractions import Fraction
 from functools import reduce
+from collections import Sequence
+
+import attr
 
 from hypothesis.errors import InvalidArgument, ResolutionFailed
-from hypothesis.control import note, assume, cleanup, current_build_context
+from hypothesis.control import note, assume, reject, cleanup, \
+    current_build_context
 from hypothesis._settings import note_deprecation
 from hypothesis.internal.cache import LRUReusedCache
 from hypothesis.searchstrategy import SearchStrategy, check_strategy
 from hypothesis.internal.compat import gcd, ceil, floor, hrange, \
-    text_type, get_type_hints, getfullargspec, implements_iterator
-from hypothesis.internal.floats import next_up, next_down, is_negative, \
-    float_to_int, int_to_float, count_between_floats
+    string_types, get_type_hints, getfullargspec, typing_root_type, \
+    implements_iterator
+from hypothesis.internal.floats import next_up, float_of, next_down, \
+    is_negative, float_to_int, int_to_float, count_between_floats
 from hypothesis.internal.charmap import as_general_categories
 from hypothesis.internal.cathetus import cathetus
 from hypothesis.internal.renaming import renamed_arguments
 from hypothesis.utils.conventions import infer, not_set
 from hypothesis.internal.reflection import proxies, required_args, \
-    define_function_signature
+    is_typed_named_tuple, define_function_signature
 from hypothesis.internal.validation import check_type, try_convert, \
     check_valid_size, check_valid_bound, check_valid_sizes, \
     check_valid_integer, check_valid_interval, check_valid_magnitude
@@ -74,19 +81,23 @@ try:
 except ImportError:
     pass
 
+try:
+    import numpy
+except ImportError:
+    numpy = None
+
 if False:
     from types import ModuleType  # noqa
     from random import Random  # noqa
     from typing import Any, Dict, Union, Sequence, Callable, Pattern  # noqa
     from typing import TypeVar, Tuple, List, Set, FrozenSet, overload  # noqa
-    from typing import Type, Mapping, Text, AnyStr, Optional  # noqa
+    from typing import Type, Text, AnyStr, Optional  # noqa
 
     from hypothesis.utils.conventions import InferType  # noqa
     from hypothesis.searchstrategy.strategies import T, Ex  # noqa
     K, V = TypeVar['K'], TypeVar['V']
     # See https://github.com/python/mypy/issues/3186 - numbers.Real is wrong!
     Real = Union[int, float, Fraction, Decimal]
-    ExtendFunc = Callable[[SearchStrategy[Union[T, Ex]]], SearchStrategy[T]]
 else:
     def overload(f):
         return f
@@ -108,7 +119,7 @@ __all__ = [
     'recursive', 'composite',
     'shared', 'runner', 'data',
     'deferred',
-    'from_type', 'register_type_strategy',
+    'from_type', 'register_type_strategy', 'emails',
 ]
 
 _strategies = set()
@@ -259,10 +270,23 @@ def none():
     return just(None)
 
 
-def one_of(
-    *args  # type: Union[SearchStrategy[Ex], Sequence[SearchStrategy[Ex]]]
-):
-    # type: (...) -> SearchStrategy[Ex]
+@overload
+def one_of(args):
+    # type: (Sequence[SearchStrategy[Any]]) -> SearchStrategy[Any]
+    pass  # pragma: no cover
+
+
+@overload
+def one_of(*args):
+    # type: (SearchStrategy[Any]) -> SearchStrategy[Any]
+    pass  # pragma: no cover
+
+
+def one_of(*args):
+    # Mypy workaround alert:  Any is too loose above; the return paramater
+    # should be the union of the input parameters.  Unfortunately, Mypy <=0.600
+    # raises errors due to incompatible inputs instead.  See #1270 for links.
+    # v0.610 doesn't error; it gets inference wrong for 2+ arguments instead.
     """Return a strategy which generates values from any of the argument
     strategies.
 
@@ -355,8 +379,13 @@ def booleans():
 @cacheable
 @defines_strategy_with_reusable_values
 def floats(
-    min_value=None, max_value=None, allow_nan=None, allow_infinity=None
+    min_value=None,  # type: Real
+    max_value=None,  # type: Real
+    allow_nan=None,  # type: bool
+    allow_infinity=None,  # type: bool
+    width=64,  # type: int
 ):
+    # type: (...) -> SearchStrategy[float]
     """Returns a strategy which generates floats.
 
     - If min_value is not None, all values will be >= min_value.
@@ -369,6 +398,13 @@ def floats(
     Where not explicitly ruled out by the bounds, all of infinity, -infinity
     and NaN are possible values generated by this strategy.
 
+    The width argument specifies the maximum number of bits of precision
+    required to represent the generated float. Valid values are 16, 32, or 64.
+    Passing ``width=32`` will still use the builtin 64-bit ``float`` class,
+    but always for values which can be exactly represented as a 32-bit float.
+    Half-precision floats (``width=16``) are only supported on Python 3.6, or
+    if Numpy is installed.
+
     Examples from this strategy have a complicated and hard to explain
     shrinking behaviour, but it tries to improve "human readability". Finite
     numbers will be preferred to infinity and infinity will be preferred to
@@ -384,14 +420,26 @@ def floats(
                     allow_nan
                 ))
 
+    if width not in (16, 32, 64):
+        raise InvalidArgument(
+            'Got width=%r, but the only valid values are the integers 16, '
+            '32, and 64.' % (width,)
+        )
+    if width == 16 and sys.version_info[:2] < (3, 6) and numpy is None:
+        raise InvalidArgument(  # pragma: no cover
+            'width=16 requires either Numpy, or Python >= 3.6'
+        )
+
     check_valid_bound(min_value, 'min_value')
     check_valid_bound(max_value, 'max_value')
 
     min_arg, max_arg = min_value, max_value
     if min_value is not None:
-        min_value = float(min_value)
+        min_value = float_of(min_value, width)
+        assert isinstance(min_value, float)
     if max_value is not None:
-        max_value = float(max_value)
+        max_value = float_of(max_value, width)
+        assert isinstance(max_value, float)
 
     check_valid_interval(min_value, max_value, 'min_value', 'max_value')
     if min_value == float(u'-inf'):
@@ -399,16 +447,17 @@ def floats(
     if max_value == float(u'inf'):
         max_value = None
 
-    if min_value is not None and min_value < min_arg:
-        min_value = next_up(min_value)
-        assert min_value > min_arg
-    if max_value is not None and max_value > max_arg:
-        max_value = next_down(max_value)
-        assert max_value < max_arg
-    if None not in (min_value, max_value) and min_value > max_value:
+    if min_value is not None and min_arg is not None and min_value < min_arg:
+        min_value = next_up(min_value, width)
+        assert min_value > min_arg  # type: ignore
+    if max_value is not None and max_arg is not None and max_value > max_arg:
+        max_value = next_down(max_value, width)
+        assert max_value < max_arg  # type: ignore
+    if min_value is not None and max_value is not None and \
+            min_value > max_value:  # type: ignore
         raise InvalidArgument(
-            'There are no floating-point values between min_value=%r and '
-            'max_value=%r' % (min_arg, max_arg))
+            'There are no %s-bit floating-point values between min_value=%r '
+            'and max_value=%r' % (width, min_arg, max_arg))
 
     if allow_infinity is None:
         allow_infinity = bool(min_value is None or max_value is None)
@@ -421,61 +470,71 @@ def floats(
                 ))
 
     if min_value is None and max_value is None:
-        return FloatStrategy(
+        result = FloatStrategy(
             allow_infinity=allow_infinity, allow_nan=allow_nan,
-        )
+        )  # type: SearchStrategy[float]
     elif min_value is not None and max_value is not None:
         if min_value == max_value:
-            return just(min_value)
+            assert isinstance(min_value, float)
+            result = just(min_value)
         elif is_negative(min_value):
             if is_negative(max_value):
-                return floats(min_value=-max_value, max_value=-min_value).map(
-                    operator.neg
-                )
+                result = floats(min_value=-max_value, max_value=-min_value)\
+                    .map(operator.neg)
             else:
-                return floats(min_value=0.0, max_value=max_value) | floats(
+                result = floats(min_value=0.0, max_value=max_value) | floats(
                     min_value=0.0, max_value=-min_value).map(operator.neg)
         elif count_between_floats(min_value, max_value) > 1000:
-            return FixedBoundedFloatStrategy(
+            result = FixedBoundedFloatStrategy(
                 lower_bound=min_value, upper_bound=max_value
             )
         else:
-            ub_int = float_to_int(max_value)
-            lb_int = float_to_int(min_value)
+            ub_int = float_to_int(max_value, width)
+            lb_int = float_to_int(min_value, width)
             assert lb_int <= ub_int
-            return integers(min_value=lb_int, max_value=ub_int).map(
-                int_to_float
+            result = integers(min_value=lb_int, max_value=ub_int).map(
+                lambda x: int_to_float(x, width)
             )
     elif min_value is not None:
+        assert isinstance(min_value, float)
         if min_value < 0:
             result = floats(
-                min_value=0.0
+                min_value=0.0, allow_infinity=allow_infinity
             ) | floats(min_value=min_value, max_value=-0.0)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
-                    lambda x: assume(not math.isnan(x)) and min_value + abs(x)
+                    lambda x: assume(not math.isnan(x)) and
+                    min_value + abs(x)  # type: ignore
                 )
             )
         if min_value == 0 and not is_negative(min_value):
             result = result.filter(lambda x: math.copysign(1.0, x) == 1)
-        return result
     else:
-        assert max_value is not None
+        assert isinstance(max_value, float)
         if max_value > 0:
             result = floats(
                 min_value=0.0,
                 max_value=max_value,
-            ) | floats(max_value=-0.0)
+            ) | floats(max_value=-0.0, allow_infinity=allow_infinity)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
-                    lambda x: assume(not math.isnan(x)) and max_value - abs(x)
+                    lambda x: assume(not math.isnan(x)) and
+                    max_value - abs(x)  # type: ignore
                 )
             )
         if max_value == 0 and is_negative(max_value):
             result = result.filter(is_negative)
-        return result
+
+    if width < 64:
+        def downcast(x):
+            try:
+                return float_of(x, width)
+            except OverflowError:
+                reject()
+        return result.map(downcast)
+    return result
 
 
 @cacheable
@@ -703,7 +762,7 @@ def iterables(elements=None, min_size=None, average_size=None, max_size=None,
 
 @defines_strategy
 def fixed_dictionaries(
-    mapping  # type: Mapping[T, SearchStrategy[Ex]]
+    mapping  # type: Dict[T, SearchStrategy[Ex]]
 ):
     # type: (...) -> SearchStrategy[Dict[T, Ex]]
     """Generates a dictionary of the same type as mapping with a fixed set of
@@ -728,12 +787,14 @@ def fixed_dictionaries(
 def dictionaries(
     keys,  # type: SearchStrategy[Ex]
     values,  # type: SearchStrategy[T]
-    dict_class=dict,  # type: Type[Mapping]
+    dict_class=dict,  # type: type
     min_size=None,  # type: int
     average_size=None,  # type: int
     max_size=None,  # type: int
 ):
-    # type: (...) -> SearchStrategy[Mapping[Ex, T]]
+    # type: (...) -> SearchStrategy[Dict[Ex, T]]
+    # Describing the exact dict_class to Mypy drops the key and value types,
+    # so we report Dict[K, V] instead of Mapping[Any, Any] for now.  Sorry!
     """Generates dictionaries of type dict_class with keys drawn from the keys
     argument and values drawn from the values argument.
 
@@ -874,7 +935,7 @@ def characters(
 @cacheable
 @defines_strategy_with_reusable_values
 def text(
-    alphabet=None,  # type: Union[Text, SearchStrategy[Text]]
+    alphabet=None,  # type: Union[Sequence[Text], SearchStrategy[Text]]
     min_size=None,   # type: int
     average_size=None,   # type: int
     max_size=None  # type: int
@@ -909,7 +970,31 @@ def text(
     elif isinstance(alphabet, SearchStrategy):
         char_strategy = alphabet
     else:
-        char_strategy = sampled_from(list(map(text_type, alphabet)))
+        if not isinstance(alphabet, Sequence):
+            note_deprecation(
+                'alphabet must be an ordered sequence, or tests may be '
+                'flaky and shrinking weaker, but a %r is not a type of '
+                'sequence.  This will be an error in future.'
+                % (type(alphabet),)
+            )
+        alphabet = list(alphabet)
+        non_string = [c for c in alphabet if not isinstance(c, string_types)]
+        if non_string:
+            note_deprecation(
+                'The following elements in alphabet are not unicode '
+                'strings, which will be an error in future:  %r'
+                % (non_string,)
+            )
+            alphabet = [str(c) for c in alphabet]
+        not_one_char = [c for c in alphabet
+                        if isinstance(c, string_types) and len(c) != 1]
+        if not_one_char:
+            note_deprecation(
+                'The following elements in alphabet are not of length '
+                'one, which leads to violation of size constraints and '
+                'will be an error in future:  %r' % (not_one_char,)
+            )
+        char_strategy = sampled_from(alphabet)
     return StringStrategy(lists(
         char_strategy, min_size=min_size, max_size=max_size
     ))
@@ -1047,6 +1132,10 @@ def builds(
     value :const:`hypothesis.infer` as a keyword argument to
     builds, instead of a strategy for that argument to the callable.
 
+    If the callable is a class defined with :pypi:`attrs`, missing required
+    arguments will be inferred from the attribute on a best-effort basis,
+    e.g. by checking :ref:`attrs standard validators <attrs:api_validators>`.
+
     Examples from this strategy shrink by shrinking the argument values to
     the callable.
     """
@@ -1071,16 +1160,28 @@ def builds(
         # Avoid an implementation nightmare juggling tuples and worse things
         raise InvalidArgument('infer was passed as a positional argument to '
                               'builds(), but is only allowed as a keyword arg')
-    hints = get_type_hints(target.__init__ if isclass(target) else target)
-    for kw in [k for k, v in kwargs.items() if v is infer]:
-        if kw not in hints:
-            raise InvalidArgument(
-                'passed %s=infer for %s, but %s has no type annotation'
-                % (kw, target.__name__, kw))
-        kwargs[kw] = from_type(hints[kw])
     required = required_args(target, args, kwargs)
-    for ms in set(hints) & (required or set()):
-        kwargs[ms] = from_type(hints[ms])
+    to_infer = set(k for k, v in kwargs.items() if v is infer)
+    if required or to_infer:
+        if isclass(target) and attr.has(target):
+            # Use our custom introspection for attrs classes
+            from hypothesis.searchstrategy.attrs import from_attrs
+            return from_attrs(target, args, kwargs, required | to_infer)
+        # Otherwise, try using type hints
+        if isclass(target):
+            if is_typed_named_tuple(target):
+                # Special handling for typing.NamedTuple
+                hints = target._field_types
+            else:
+                hints = get_type_hints(target.__init__)
+        else:
+            hints = get_type_hints(target)
+        if to_infer - set(hints):
+            raise InvalidArgument(
+                'passed infer for %s, but there is no type annotation'
+                % (', '.join(sorted(to_infer - set(hints)))))
+        for kw in set(hints) & (required | to_infer):
+            kwargs[kw] = from_type(hints[kw])
     # Mypy doesn't realise that `infer` is gone from kwargs now
     kwarg_strat = fixed_dictionaries(kwargs)  # type: ignore
     return tuples(tuples(*args), kwarg_strat).map(
@@ -1131,7 +1232,6 @@ def from_type(thing):
     from hypothesis.searchstrategy import types
 
     if typing is not None:  # pragma: no branch
-        fr = typing._ForwardRef
         if not isinstance(thing, type):
             # At runtime, `typing.NewType` returns an identity function rather
             # than an actual type, but we can check that for a possible match
@@ -1148,12 +1248,13 @@ def from_type(thing):
                 return one_of([from_type(t) for t in args])
         # We can't resolve forward references, and under Python 3.5 (only)
         # a forward reference is an instance of type.  Hence, explicit check:
-        elif type(thing) == fr:  # pragma: no cover
+        elif hasattr(typing, '_ForwardRef') and \
+                type(thing) == typing._ForwardRef:  # pragma: no cover
             raise ResolutionFailed(
                 'thing=%s cannot be resolved.  Upgrading to python>=3.6 may '
                 'fix this problem via improvements to the typing module.'
                 % (thing,))
-    if not isinstance(thing, type):
+    if not types.is_a_type(thing):
         raise InvalidArgument('thing=%s must be a type' % (thing,))
     # Now that we know `thing` is a type, the first step is to check for an
     # explicitly registered strategy.  This is the best (and hopefully most
@@ -1163,27 +1264,27 @@ def from_type(thing):
     if thing in types._global_type_lookup:
         strategy = types._global_type_lookup[thing]
         if not isinstance(strategy, SearchStrategy):
-            strategy = strategy(thing)
-        if strategy.is_empty:
+            strategy = strategy(thing)  # type: ignore
+        if strategy.is_empty:  # type: ignore
             raise ResolutionFailed(
                 'Error: %r resolved to an empty strategy' % (thing,))
-        return strategy
+        return strategy  # type: ignore
     # If there's no explicitly registered strategy, maybe a subtype of thing
     # is registered - if so, we can resolve it to the subclass strategy.
     # We'll start by checking if thing is from from the typing module,
     # because there are several special cases that don't play well with
     # subclass and instance checks.
     if typing is not None:  # pragma: no branch
-        if isinstance(thing, typing.TypingMeta):
+        if isinstance(thing, typing_root_type):
             return types.from_typing_type(thing)
     # If it's not from the typing module, we get all registered types that are
     # a subclass of `thing` and are not themselves a subtype of any other such
     # type.  For example, `Number -> integers() | floats()`, but bools() is
     # not included because bool is a subclass of int as well as Number.
     strategies = [
-        v if isinstance(v, SearchStrategy) else v(thing)
+        v if isinstance(v, SearchStrategy) else v(thing)  # type: ignore
         for k, v in types._global_type_lookup.items()
-        if issubclass(k, thing) and sum(
+        if isinstance(k, type) and issubclass(k, thing) and sum(
             types.try_issubclass(k, typ) for typ in types._global_type_lookup
         ) == 1
     ]
@@ -1196,23 +1297,22 @@ def from_type(thing):
         return one_of(strategies)
     # If we don't have a strategy registered for this type or any subtype, we
     # may be able to fall back on type annotations.
-    # Types created via typing.NamedTuple use a custom attribute instead -
-    # but we can still use builds(), if we work out the right kwargs.
-    if issubclass(thing, tuple) and hasattr(thing, '_fields') \
-            and hasattr(thing, '_field_types'):
-        kwargs = {k: from_type(thing._field_types[k]) for k in thing._fields}
-        return builds(thing, **kwargs)
     if issubclass(thing, enum.Enum):
         assert len(thing), repr(thing) + ' has no members to sample'
         return sampled_from(thing)
-    # If the constructor has an annotation for every required argument,
-    # we can (and do) use builds() without supplying additional arguments.
+    # If we know that builds(thing) will fail, give a better error message
     required = required_args(thing)
-    if not required or required.issubset(get_type_hints(thing.__init__)):
-        return builds(thing)
-    # We have utterly failed, and might as well say so now.
-    raise ResolutionFailed('Could not resolve %r to a strategy; consider '
-                           'using register_type_strategy' % (thing,))
+    if not any([
+        not required,
+        required.issubset(get_type_hints(thing.__init__)),
+        attr.has(thing),
+        # NamedTuples are weird enough that we need a specific check for them.
+        is_typed_named_tuple(thing),
+    ]):
+        raise ResolutionFailed('Could not resolve %r to a strategy; consider '
+                               'using register_type_strategy' % (thing,))
+    # Finally, try to build an instance by calling the type object
+    return builds(thing)
 
 
 @cacheable
@@ -1439,7 +1539,7 @@ def decimals(
 
 def recursive(
     base,  # type: SearchStrategy[Ex]
-    extend,  # type: ExtendFunc
+    extend,  # type: Callable[[SearchStrategy[Any]], SearchStrategy[T]]
     max_leaves=100,  # type: int
 ):
     # type: (...) -> SearchStrategy[Union[T, Ex]]
@@ -1996,8 +2096,11 @@ def data():
     return DataStrategy()
 
 
-def register_type_strategy(custom_type, strategy):
-    # type: (type, Union[type, Callable[[type], SearchStrategy]]) -> None
+def register_type_strategy(
+    custom_type,  # type: type
+    strategy,  # type: Union[SearchStrategy, Callable[[type], SearchStrategy]]
+):
+    # type: (...) -> None
     """Add an entry to the global type-to-strategy lookup.
 
     This lookup is used in :func:`~hypothesis.strategies.builds` and
@@ -2015,7 +2118,7 @@ def register_type_strategy(custom_type, strategy):
     # TODO: We would like to move this to the top level, but pending some major
     # refactoring it's hard to do without creating circular imports.
     from hypothesis.searchstrategy import types
-    if not isinstance(custom_type, type):
+    if not types.is_a_type(custom_type):
         raise InvalidArgument('custom_type=%r must be a type')
     elif not (isinstance(strategy, SearchStrategy) or callable(strategy)):
         raise InvalidArgument(
@@ -2062,4 +2165,22 @@ def deferred(definition):
     return DeferredStrategy(definition)
 
 
+@defines_strategy_with_reusable_values
+def emails():
+    """A strategy for generating email addresses as unicode strings. The
+    address format is specific in :rfc:`5322#section-3.4.1`. Values shrink
+    towards shorter local-parts and host domains.
+
+    This strategy is useful for generating "user data" for tests, as
+    mishandling of email addresses is a common source of bugs. Future
+    updates will generate more complicated addresses allowed by the RFC.
+    """
+    from hypothesis.provisional import domains
+    local_chars = string.ascii_letters + string.digits + "!#$%&'*+-/=^_`{|}~"
+    local_part = text(local_chars, min_size=1, max_size=64)
+    # TODO: include dot-atoms, quoted strings, escaped chars, etc in local part
+    return builds(u'{}@{}'.format, local_part, domains()).filter(
+        lambda addr: len(addr) <= 255)
+
+
 assert _strategies.issubset(set(__all__)), _strategies - set(__all__)
